# History Mar 11 2014 Add new functions to handle multiple genes

# Main function
treat <- function(formula, data, snp.vars, op=NULL) {

  op <- default.list(op, 
         c("two.way", "backward", "cal.pvalue", "path.op", "path.op.back",
           "indicator2", "thr.sample", "thr.grp.sample", "thr.leaf.size",
           "nleaf", "cut.point", "nperm", "trace", "seed", "debug", "bin",
           "max.missRate", "check.platform"),
           list(FALSE, TRUE, TRUE, NULL, NULL,
                NULL, 50, 2, 20,
                5, NULL, 10000, FALSE, 0, FALSE, TRUE,
                0.2, 1),
         checkList=list(0:1, 0:1, 0:1, NA, NA,
                        NA, NA, NA, NA,
                        NA, NA, NA, 0:1, NA, 0:1, 0:1,
                        NA, 0:1))

  
  if (op$check.platform) {
    cat("NOTE: The TREAT package is ONLY for 64-bit operating systems\n")
    os   <- tolower(.Platform$OS.type)
    arch <- .Platform$r_arch
    temp <- nchar(arch)
    if (!temp) cat("NOTE: Could not determine the OS architecture\n") 
    if ((os == "windows") && (arch != "x64")) {
      stop("ERROR: The TREAT package is ONLY for 64-bit operating systems")
    }
  }

  if (!is.data.frame(data)) stop("data must be a data frame")
  if (is.null(dim(data))) stop("data has no dimension")
  nr <- nrow(data) 
  if (is.null(rownames(data))) rownames(data) <- paste("row", 1:nr, sep="")
  if (is.null(colnames(data))) colnames(data) <- paste("V", 1:ncol(data), sep="")
  cnames <- colnames(data)  
  if (!("formula" %in% class(formula))) stop("formula must be a formula")
   
  rm.snps.coding   <- NULL
  rm.snps.missRate <- NULL

  vars         <- all.vars(formula)
  nvars        <- length(vars)
  response.var <- vars[1]
  covars       <- NULL
  if (nvars > 1) covars <- vars[2:nvars]
  covars       <- get_colNames(cnames, covars)
  ncovars      <- length(covars)
  snp.vars     <- get_colNames(cnames, snp.vars)
  nsnps        <- length(snp.vars)
  snpFlag      <- nsnps
  if (nsnps) {
    temp <- snp.vars %in% vars
    if (any(temp)) {
      str <- paste(snp.vars[temp], collapse=", ", sep="")
      str <- paste("The variables ", str, " should not be part of the formula", sep="")
      stop(str)
    } 
  } 

  if (!nsnps) {
    # Use other variables in data.
    temp     <- !(cnames %in% vars)
    snp.vars <- cnames[temp]
    nsnps    <- length(snp.vars) 
    snpFlag  <- 1
  }
  if (!nsnps) stop("No snp variables included")
  N.Marker.All <- nsnps

  # Check and recode the response
  vec <- data[, response.var]
  lev <- unique(vec)
  if (length(lev) != 2) stop("Response variable must be binary")
  temp <- !(vec %in% 0:1)
  if (any(temp)) {
    str <- paste("Response variable is not 0-1, recoding ", lev[1], " to 0, ",
                 lev[2], " to 1", sep="")
    warning(str)
    temp1      <- vec %in% lev[1]
    temp2      <- vec %in% lev[2]
    vec[temp1] <- 0
    vec[temp2] <- 1
    data[, response.var] <- vec
    rm(temp1, temp2, str)
  }
  rm(vec, temp, lev)
  gc()

  # Subset the data if needed
  x <- model.matrix(formula, data)
  if (nrow(x) != nr) {
    temp <- rownames(data) %in% rownames(x)
    data <- data[temp, , drop=FALSE] 
  }
  if (!nrow(data)) stop("All rows from data have been removed")

  # Check missing rate for snps
  if (op$max.missRate < 1) {
    temp <- check_miss(data, snp.vars, op$max.missRate)
    rm.snps.missRate <- temp$remove
    if (!is.null(rm.snps.missRate)) {
      temp     <- !(snp.vars %in% rm.snps.missRate)
      snp.vars <- snp.vars[temp]
    }
  }

  # Obtain initial covar vector
  n <- ncol(x)
  if (n) {
    covar <- x[, 1]
    if (n > 1) {
      for (i in 2:n) covar <- paste(covar, "+", x[, i], sep="")
    } 
    covar <- as.integer(as.factor(covar)) - 1
  } else {
    covar <- rep(0, nrow(data))
  }
  rm(x, temp)
  gc()

  # Check the response
  if (!(response.var %in% cnames)) stop("response.var is not found in data")
  
  temp <- !(as.numeric(data[, response.var]) %in% c(0, 1))
  if (any(temp)) stop("response.var is not coded correctly")
  nr <- nrow(data)

  # Check that the snps are 0-1-2
  mat  <- as.matrix(data[, snp.vars, drop=FALSE])
  mat  <- matrix(as.numeric(mat), nrow=nr, byrow=FALSE)
  colnames(mat) <- snp.vars
  temp <- !(mat %in% c(0:2, NA))
  dim(temp) <- dim(mat)

  temp <- colSums(temp) > 0
  if (any(temp)) {   
    rm.snps.coding <- snp.vars[temp]
    snp.vars       <- snp.vars[!temp] 
    cat("NOTE: The following SNPs were removed because they were not coded correctly:\n")
    print(rm.snps.coding)
    if (length(snp.vars)) mat <- mat[, snp.vars, drop=FALSE]
  }
  nsnps <- length(snp.vars)
  if (!nsnps) stop("All snp variables have been removed")

  # Get rows to keep
  temp <- !(mat %in% 0:2)
  dim(temp) <- dim(mat)
  keep <- rowSums(temp) == 0

  covar <- covar[keep]
  if (length(unique(covar)) > length(covar) - op$thr.grp.sample) stop("covariates define too many levels")
  if (op$debug) print(table(covar, exclude=NULL))

  # Get binary SNPs
  mat01 <- mat %in% 0:1
  mat12 <- mat %in% 1:2
  dim(mat01) <- dim(mat)
  dim(mat12) <- dim(mat)
  endpoint   <- ncol(mat01)
  mat   <- cbind(data[keep, response.var], mat01[keep, , drop=FALSE], mat12[keep, , drop=FALSE])
  rm(mat01, mat12, temp)
  gc()
 
  rm(data, covars, keep)
  gc()

  N.Marker.Test <- nsnps

  ret <- gene.ctree3(mat, covar, 
             op[["thr.sample", exact=TRUE]], op[["thr.grp.sample", exact=TRUE]],
             op$thr.leaf.size, op$nleaf, op[["cut.point", exact=TRUE]], op$nperm, 
             backward=op$backward, trace=op$trace, seed=op$seed,
             two.way=op$two.way, indicator2=op[["indicator2", exact=TRUE]],
             cal.pval=op$cal.pval, path.op=op[["path.op", exact=TRUE]], 
             path.op.back=op[["path.op.back", exact=TRUE]], 
             debug=op$debug, bin=op$bin)
  #ret$data  <- mat
  #ret$covar <- covar
  ret$op    <- op

  # 1. Id like to remove the first column in tree.model.forward, of which the entries are all 1.way. 
  # This column was used in debugging and I forgot to remove it before I sent the code to you.
  # 2. I want to make the column indi.ID more interpretable. The modification should be minor. 
  # Recall that you create a lot of indicators by assigning rules to the SNPs. 
  # For example, indicator1 = (rs1 == 0), indicator2 = (rs1 == 2), etc. 
  # What my c code knows are the indexes of these indicators, rather than their meanings. 
  # So the c code returns the indexes of indicators chosen as the splitting rules (indi.ID). 
  # Could you please write code to convert the indexes in indi.ID into some strings? 
  # For example, the example in the manual gives
  # > tr$tree.model.forward
  # 1.way indi.ID parent child1 child2
  # 1  TRUE       2      1      2      3
  # 2  TRUE       4      2      4      5
  # 3  TRUE      12      5      6      7
  # 4  TRUE      16      6      8      9
  # Suppose that the 2nd, 4th, 12th, 16th indicators (0/1) are defined by examining the 
  # conditions rs1 == 2, rs2 == 2, rs7 == 0 and rs8 = 0, respectively. 
  # I want the tree.model.forward looks like this
  # > tr$tree.model.forward
  #     rule parent child1(TRUE) child2(FALSE)
  #      rs1=2      1      2      3
  #      rs2=2      2      4      5
  #      rs7=0      5      6      7
  #      rs8=0      6      8      9
  
  tree      <- ret$tree.model.forward
  if (nrow(tree)) {
    indi      <- tree[, "indi.ID"]
    tree      <- tree[, -(1:2)]
    val       <- rep(0, length(indi))
    temp      <- (indi <= endpoint) # From mat01  SNP=2
    val[temp] <- 2
    SNPV      <- c(snp.vars, snp.vars)
    snps      <- SNPV[indi]
    rule      <- paste(snps, "=", val, sep="") 
    tree      <- cbind(rule, tree)
    colnames(tree) <- c("splitting rule", "parent", "child1 (rule = FALSE)", "child2 (rule = TRUE)")
  } else {
    tree <- NULL
  }
  ret$tree.model.forward <- tree

  tree      <- ret$tree.model.backward
  if (nrow(tree)) { 
    tree[, "merge"] <- "-->"
    tree            <- tree[, c("child1", "child2", "merge", "parent")]
  } else {
    #tree <- NULL
  }
  ret$tree.model.backward <- tree

  ret[["snps"]]             <- snp.vars
  ret[["rm.snps.coding"]]   <- rm.snps.coding
  ret[["rm.snps.missRate"]] <- rm.snps.missRate

  # Get the best snps
  tr             <- ret
  model          <- tr$tree.model.forward
  min.unadj.pval <- min(tr$unadj.pval) 

  # determine the optimal tree 
  #[1] is used to handle the case if minimum p-value is attained at more than 
  # one elements in unadj.pval 
  num.of.leaves <- tr$cut.point[ which(tr$unadj.pval == min.unadj.pval)[1] ] 

  num.of.splits <- num.of.leaves - 1 
  chosen.rule   <- as.character(model[1:num.of.splits, "splitting rule"]) 

  # remove characters after "="
  chosen.snp    <- sub("=.*", "", chosen.rule) 
 
  # remove duplicated terms but keep their original order 
  # chosen.snp is then ready to be placed in the final summary table
  ret$chosen.snp <- chosen.snp[ !duplicated(chosen.snp) ] 
  ret$N.Marker.All  <- N.Marker.All
  ret$N.Marker.Test <- N.Marker.Test

  ret

} # END: treat

# Function to get column names from vectors
get_colNames <- function(cnames, vars) {

  if (is.null(vars)) return(vars)
  vars  <- as.character(vars)
  ret   <- vars
  N     <- length(cnames)
  temp  <- vars %in% 1:N
  varsn <- vars[temp]
  varsc <- vars[!temp]
  if (length(varsc)) {
    temp <- (!(varsc %in% cnames))
    miss <- varsc[temp]
    if (length(miss)) {
      str <- paste(miss, collapse=", ", sep="")
      str <- paste("The variable(s) ", str, " were not found in data", sep="")
      stop(str)
    }
  }
  if (length(varsn)) {
    ret <- c(cnames[as.integer(varsn)], varsc)
  }  

  ret <- unique(ret)
  ret

} # END: get_colNames

# Function to assign a default value to an element in a list
default.list <- function(inList, names, default, error=NULL,
                         checkList=NULL) {

  # inList      List
  # names       Vector of names of items in inList
  # default     List of default values to assign if a name is not found
  #             The order of default must be the same as in names.
  # error       Vector of TRUE/FALSE if it is an error not to have the
  #             name in the list. 
  #             The default is NULL
  # checkList   List of valid values for each name.
  #             Use NA to skip a list element.
  #             The default is NULL

  n1 <- length(names)
  n2 <- length(default)
  if (n1 != n2) stop("ERROR: in calling default.list")

  if (is.null(error)) {
    error <- rep(0, times=n1)
  } else if (n1 != length(error)) {
    stop("ERROR: in calling default.list")
  }

  if (!is.null(checkList)) {
    if (n1 != length(checkList)) stop("ERROR: in calling default.list")
    checkFlag <- 1
  } else {
    checkFlag <- 0
  } 

  if (is.null(inList)) inList <- list()

  listNames <- names(inList)
  for (i in 1:n1) {
    if (!(names[i] %in% listNames)) {
      if (!error[i]) {
        inList[[names[i]]] <- default[[i]]
      } else {
        temp <- paste("ERROR: the name ", names[i], " was not found", sep="")
        stop(temp)
      }
    } else if (checkFlag) {
      temp <- checkList[[i]]
      if (!all(is.na(temp))) {
        if (!all(inList[[names[i]]] %in% checkList[[i]])) {
          temp <- paste("ERROR: the name '", names[i], 
                      "' has an invalid value", sep="")
          stop(temp)
        }
      }
    }
  }

  inList

} # END: default.list

# Function to check and return the missing rates
check_miss <- function(data, geno.vars, missRate) {

  remove            <- NULL
  temp              <- is.na(as.matrix(data[, geno.vars]))
  rate              <- colSums(temp)/nrow(data)
  names(rate)       <- geno.vars
  temp              <- (rate > missRate)
  temp[is.na(temp)] <- TRUE
  if (any(temp)) remove <- geno.vars[temp]
     
  list(remove=remove, missvec=rate)

} # END: check_miss

# Function to call the operating system
callOS <- function(command, intern=FALSE) {

  # Determine the platform
  os      <- .Platform$OS.type
  winFlag <- (os == "windows")

  if (winFlag) {
    ret <- shell(command, intern=intern)
  } else {
    ret <- system(command, intern=intern)
  }
  ret

} # END: callOS


# Function to scan a file
scan.file <- function(f, header=0, sep="") {

  row1 <- scan(f, what="character", sep=sep, nlines=1)
  nc   <- length(row1)
  skip <- 1
  if (!header) skip <- 0
  x    <- matrix(scan(f, what="character", skip=skip, sep=sep), byrow=TRUE, ncol=nc)
  if (header) colnames(x) <- row1

  x

} # END: scan.file

# Function to get gene range
TR.get.range <- function(gene.names, hg.file, delta=20000) {

  temp <- load(hg.file)
  if (temp != "gene.data") {
    gene.data <- eval(parse(text=temp))
    eval(parse(text=paste("rm(", temp, ")", sep="")))
    temp <- gc(verbose=FALSE)
  }

  temp <- gene.data[, "Gene"] %in% gene.names
  if (!any(temp)) stop("ERROR: None of the gene names were found in the data base")
  gene.data <- gene.data[temp, , drop=FALSE]
  miss <- !(gene.names %in% gene.data[, "Gene"])
  if (any(miss)) {
    print(miss)
    print("NOTE: The above gene names were not found in the data base")
  }
  gene.data <- as.data.frame(gene.data, stringsAsFactors=FALSE)
  gene.data[, "Start"] <- as.numeric(gene.data[, "Start"]) - delta
  gene.data[, "Stop"]  <- as.numeric(gene.data[, "Stop"]) + delta

  gene.data

} # END: TR.get.range

checkForSep <- function(dir) {

  if (is.null(dir)) return("")

  if (dir %in% c("", " ", "  ")) {
    dir <- paste(".", .Platform$file.sep, sep="")
    return("")
  }

  n <- nchar(dir)
  if (substring(dir, n, n) %in% c("\\", "/")) {
    return(dir)
  } else {
    dir <- paste(dir, .Platform$file.sep, sep="")
    return(dir)
  }

} # END: checkForSep

# Function to get the genotype data
treat.getGeno <- function(plink.files, snps) {

  m <- length(plink.files)
  if (m == 3) {
    ret    <- read.plink(plink.files[1], plink.files[2], plink.files[3], select.snps=snps) 
    fam    <- ret$fam
    ids    <- paste(fam[, 1], ".", fam[, 2], sep="")
    gmat   <- ret$genotypes  
    gmat   <- gmat@.Data
    cnames <- colnames(gmat) 
    nc     <- length(cnames)

    # Genotypes are coded 1,2,3 with 0=missing
    gmat           <- matrix(as.integer(gmat), byrow=FALSE, ncol=nc)
    temp           <- gmat %in% 0
    gmat[temp]     <- NA
    gmat           <- 3 - gmat
    rownames(gmat) <- ids
    colnames(gmat) <- cnames
  }
  gmat <- as.data.frame(gmat, stringsAsFactors=FALSE)

  gmat

} # END: treat.getGeno

# Function to return snps to be included in analysis
treat.getSnps <- function(gene.obj, map.file) {

  x     <- scan.file(map.file, header=0)
  ngene <- nrow(gene.obj)
  g.chr <- gene.obj[, "Chr"]
  g.a   <- gene.obj[, "Start"]
  g.b   <- gene.obj[, "Stop"]
  g.g   <- gene.obj[, "Gene"]
  m.chr <- x[, 1]
  m.snp <- x[, 2]
  m.loc <- as.numeric(x[, 4])
  rm(x)
  gc()

  flag     <- 0
  snp.list <- list()
  for (i in 1:ngene) {
    temp <- (m.chr %in% g.chr[i]) & (m.loc >= g.a[i]) & (m.loc <= g.b[i])
    temp[is.na(temp)] <- FALSE
    if (any(temp)) {
      snp.list[[g.g[i]]] <- m.snp[temp]
      flag <- 1
    } 
  }

  if (!flag) stop("ERROR: no SNPs to include")

  snp.list

} # END: treat.getSnps

# Function to get the data frame with snps for a gene
treat.get.gdata <- function(plink.files, snps, data, id.var, gmat) {

  # Get genotype data  
  if (is.null(gmat)) gmat <- treat.getGeno(plink.files, snps)

  # Subset data and gmat
  ids <- intersect(rownames(gmat), data[, id.var])
  if (!length(ids)) stop("ERROR: no intersecting subject ids")
  temp <- data[, id.var] %in% ids
  data <- data[temp, , drop=FALSE]
  temp <- rownames(gmat) %in% ids
  gmat <- gmat[temp, , drop=FALSE]

  # Order data by gmat
  temp <- match(rownames(gmat), data[, id.var])
  data <- data[temp, , drop=FALSE]

  # Merge
  data <- cbind(data, gmat)

  data

} # END: treat.get.gdata

# Function to return best snps


# Main function for plink files 
treat.plink <- function(plink.files, formula, data, id.vars, 
              gene.obj, op=NULL) {

  # plink.files  c(bed, bim, fam) or (ped, map)
  # gene.obj  Vector of gene names or matrix with cols Chr, Start, Stop
  # op
  # out.dir   Defaults to getwd()
  # gene.db

  op <- default.list(op, c("delta", "gene.db", "M"), list(20000, "hg18", 1000))
  
  if (!is.data.frame(data)) stop("data must be a data frame")
  if (is.null(dim(data))) stop("data has no dimension")
  if (!("formula" %in% class(formula))) stop("formula must be a formula")


  delta   <- floor(op$delta)
  out.dir <- op[["out.dir", exact=TRUE]]
  if (is.null(out.dir)) out.dir <- getwd()
  out.dir <- checkForSep(out.dir)
  gene.db <- op$gene.db

  id.vars <- unique(id.vars)
  if (length(id.vars) != 2) stop("ERROR: id.vars must be a character vector of length 2")
  temp <- !(id.vars %in% colnames(data))
  if (any(temp)) {
    miss <- paste(id.vars[temp], collapse=", ", sep="")
    temp <- paste("ERROR: the id variables ", miss, " were not found in data", sep="")
    stop(temp) 
  }

  plink.files <- unique(plink.files)
  nplink      <- length(plink.files)
  if (!(nplink %in% 2:3)) stop("ERROR: plink.files must be a character vector of length 2 or 3")

  # Get gene info
  loc.flag <- nrow(gene.obj)
  if (is.null(loc.flag)) loc.flag <- 0
  
  # If locations were not given, we need the gene data base
  if (!loc.flag) {
    # Check gene data base
    hg.flag <- 1
    if (gene.db == "hg19") {
      gene.file <- system.file("sampleData", "hg19.rda", package="TREAT", 
                         lib.loc=op[["lib.loc", exact=TRUE]], mustWork=TRUE)
    } else if (gene.db == "hg18") {
      gene.file <- system.file("sampleData", "hg18.rda", package="TREAT", 
                         lib.loc=op[["lib.loc", exact=TRUE]], mustWork=TRUE)
    } else {
      gene.file <- gene.db
      hg.flag   <- 0
    }
    if (!file.exists(gene.file)) stop(paste("ERROR: ", gene.file, " not found", sep=""))
  
    # Gene names
    gene.obj <- TR.get.range(gene.obj, gene.file, delta=delta)
  } else {
    # Check object and add/subtract delta
    gene.obj <- as.data.frame(gene.obj, stringsAsFactors=FALSE)
    cnames   <- c("Chr", "Start", "Stop")
    temp     <- !(cnames %in% colnames(gene.obj))
    if (any(temp)) {
      miss <- paste(cnames[temp], collapse=", ", sep="")
      temp <- paste("The columns ", miss, " are missing in gene.obj", sep="")
      stop(temp)
    }
    gene.obj[, "Start"] <- as.numeric(gene.obj[, "Start"]) - delta
    gene.obj[, "Stop"]  <- as.numeric(gene.obj[, "Stop"]) + delta
  }
  n.gene <- nrow(gene.obj)
  if (!("Gene" %in% colnames(gene.obj))) {
    gene.obj[, "Gene"] <- paste("Gene", 1:n.gene, sep="") 
  }

  # Keep only the variables we need in the data frame
  vars <- unique(c(id.vars, all.vars(formula)))
  data <- data[, vars, drop=FALSE]
  
  # Create new id variable for id.vars[1]
  id.var <- id.vars[1]
  data[, id.var] <- paste(data[, id.var], ".", data[, id.vars[2]], sep="")

  # Get the snps to include
  snp.list <- treat.getSnps(gene.obj, plink.files[2])
  all.snps <- unique(as.vector(unlist(snp.list)))
  M.snp    <- length(all.snps)
  M.flag   <- M.snp <= op$M

  # Get genotype data merged with phenotype if there are few snps 
  if (M.flag) data <- treat.get.gdata(plink.files, all.snps, data, id.var, NULL)

  rm(all.snps)
  gc()

  # Summary data frame
  ret.sum <- data.frame(gene.obj[, "Gene"], stringsAsFactors=FALSE)
  colnames(ret.sum)          <- "Gene"
  ret.sum[, "Pvalue"]        <- NA
  ret.sum[, "N.Marker.All"]  <- NA
  ret.sum[, "N.Marker.Test"] <- NA 
  ret.sum[, "Best.SNPs"]     <- ""

  # Loop over each gene
  out.files <- rep("ERROR", n.gene)
  names(out.files) <- gene.obj[, "Gene"]
  for (i in 1:n.gene) {
    chr   <- gene.obj[i, "Chr"]
    start <- gene.obj[i, "Start"]
    stop  <- gene.obj[i, "Stop"]
    gene  <- gene.obj[i, "Gene"]
    snps  <- snp.list[[gene]]

    if (!M.flag) {
      data2 <- treat.get.gdata(plink.files, snps, data, id.var, NULL)
    } else {
      data2 <- data
    }

    obj   <- try(treat(formula, data2, snps, op=op), silent=TRUE)
    if ("try-error" %in% class(obj)) {
      print(obj)
    } else {
      out   <- paste(out.dir, gene, ".rda", sep="")
      save(obj, file=out)
      out.files[i] <- out

      ret.sum[i, "Pvalue"]        <- obj$adj.pval
      ret.sum[i, "N.Marker.All"]  <- obj$N.Marker.All
      ret.sum[i, "N.Marker.Test"] <- obj$N.Marker.Test
      ret.sum[i, "Best.SNPs"]     <- paste(obj$chosen.snp, collapse=",", sep="")
    }

    rm(data2, snps, obj)
    gc()
  }  

  list(summary=ret.sum, saved.files=out.files)

} # END: treat.plink




