#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include <R.h>
#include <R_ext/Memory.h>
#include <Rmath.h>
#include <Rinternals.h>
#include <R_ext/Rdynload.h>


#define ALMOST_ZERO 1e-16
#define NUM_ZERO 1e-100
#define ERROR_SINGULAR_MATRIX 1


/* Function to allocate memory for a double vector */
double * dVec_alloc(int n, int initFlag, double initVal)
{
  int i;
  double *ret, *p;

  ret = (double *) R_Calloc(n, double);
  if (initFlag) {
    for (i=0, p=ret; i<n; i++, p++) *p = initVal;
  }
  
  return(ret);

} /* END: dVec_alloc */

/* Function to allocate an integer matrix */
double ** dMat_alloc(int nrow, int ncol, int initFlag, double initVal)
{
  double **mat, **ptr;
  int i;

  mat = (double **) R_Calloc(nrow, double *);
  for (i=0, ptr=mat; i<nrow; i++, ptr++) *ptr = dVec_alloc(ncol, initFlag, initVal);

  return(mat);

} /* END: dMat_alloc */

/* Function to free a matrix */
void matrix_free(void **x, int n)
{
  int i;
  for (i=0; i<n; i++) {
    if (x[i]) R_Free(x[i]);
  }
  R_Free(x);

} /* END: matrix_free */


/* Multiply to matrices (matrix mult)  */
void matrixMult(double **m1, int m1_nr, int m1_nc, double **m2, int m2_nc, double **ret)
{
  int i, j, k;
  double sum;

  for (i=0; i<m1_nr; i++) {
    for (j=0; j<m2_nc; j++) {
      sum = 0.;
      for (k=0; k<m1_nc; k++) sum += m1[i][k]*m2[k][j]; 
      ret[i][j] = sum;
    }
  }

} /* END: matrixMult */

/* Factor a symmetric positive definite matrix */
static int symPosFactor(double **mat, int n, double **ret, double *retdiag)
{
  int i, j, k;
  double sum, save, *ptr;
  
  /* Copy mat to ret */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[i][j] = mat[i][j];
  }

  for (i=0, ptr=retdiag; i<n; i++, ptr++) {
    for (j=i; j<n; j++) {
      sum = ret[i][j];
      for (k=i-1; k>-1; k--) sum -= ret[i][k]*ret[j][k];
      if (i == j) {
        if (sum < NUM_ZERO) return(ERROR_SINGULAR_MATRIX); 
        save = sqrt(sum);
        *ptr = save;
      } else {
        ret[j][i] = sum/save;
      }
    }
  }

  /* Zero out the diagonal and above */
  for (i=0; i<n; i++) {
    for (j=i; j<n; j++) ret[i][j] = 0.;
  }

  return(0);

} /* END: symPosFactor */

/* Invert a factor */
static void symPosFacInv(double **L, double *diag, int n, double **ret)
{
  int i, j, k;
  double sum;
  
  /* Copy L to ret */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[i][j] = L[i][j];
  }

  for (i=0; i<n; i++) {
    ret[i][i] = 1./diag[i];
    for (j=i+1; j<n; j++) {
      sum = 0.;
      for (k=i; k<=j-1; k++) sum -= ret[j][k]*ret[k][i];
   
      ret[j][i] = sum/diag[j];
    }
  }

} /* END: symPosFacInv */

/* Transpose of a square matrix */
static void matTranspose(double **mat, int n, double **ret)
{
  int i, j;

  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[j][i] = mat[i][j];   
  }

} /* END: rmatTranspose */

/* Inverse of symmetric positive definite matrix */
static int symPosMatInv(double **mat, int n, double **ret)
{
  double **L, **Linv, *diag;  
  int i;

  L = dMat_alloc(n, n, 0, 0);
  diag = (double *) R_Calloc(n, double);

  /* Get cholesky lower triangle */
  i = symPosFactor(mat, n, L, diag);
  if (i) {
    R_Free(diag);
    matrix_free((void **) L, n);
    return(i);
  }

  /* Get the lower inverse */
  Linv = dMat_alloc(n, n, 0, 0);
  symPosFacInv(L, diag, n, Linv);
  R_Free(diag);

  /* Get the transpose of Linv */
  matTranspose(Linv, n, L);

  /* Inverse is t(L^-1)(L^-1) */
  matrixMult(L, n, n, Linv, n, ret);

  matrix_free((void **) L, n);
  matrix_free((void **) Linv, n);

  return(0);

} /* END: symPosMatInv */

/* Function to compute the inverse of a covariance matrix */
int cov_inv(double **cov, int n, double **inv)
{
  double cc, a, b, d;
  int ret;

  switch (n) {
    case 0:
      Rprintf("\nERROR: dimension of covariance matrix is 0\n");
    case 1:
      a = cov[0][0];
      if (fabs(a) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      inv[0][0] = 1.0/a; 
      break;
    case 2:
      a  = cov[0][0];
      b  = cov[0][1];
      d  = cov[1][1];
      cc = a*d - b*b;
      if (fabs(cc) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      cc = 1.0/cc;
      inv[0][0] = d*cc;
      inv[0][1] = -b*cc;
      inv[1][0] = -b*cc;
      inv[1][1] = a*cc;
      break;
    default:
      ret = symPosMatInv(cov, n, inv);
      if (ret) return(ret);
      break; 
  } /* END: switch */

  return(0);

} /* END: cov_inv */

/* Function to invert a matrix */
static void matrixInverse(double **mat, int n, double **ret, int *error)
{
  int i, j, r, *p, *ptri, k, hi;
  double max, temp, *hv, hr;

  /* Copy mat to ret */
  for (i=0; i<n; i++) {
    for (j=0; j<n; j++) ret[i][j] = mat[i][j];
  }

  p = (int *) R_Calloc(n, int);
  hv = (double *) R_Calloc(n, double);
 
  for (j=0, ptri=p; j<n; j++, ptri++) *ptri = j; 
  for (j=0; j<n; j++) {

    /* pivot search */
    max = fabs(ret[j][j]);
    r = j;
    for (i=j+1; i<n; i++) {
      temp = fabs(ret[i][j]);
      if (temp > max) {
        max = temp;
        r = i;
      }
    }

    if (max < ALMOST_ZERO) {
      *error = ERROR_SINGULAR_MATRIX;
      return;
    }
 
    /* Row interchange */
    if (r > j) {
      for (k=0; k<n; k++) {
        hr = ret[j][k];
        ret[j][k] = ret[r][k];
        ret[r][k] = hr;
      }
      hi = p[j];
      p[j] = p[r];
      p[r] = hi; 
    }

    /* Transformation */
    hr = 1.0/ret[j][j];
    for (i=0; i<n; i++) ret[i][j] = hr*ret[i][j];
    ret[j][j] = hr;
    for (k=0; k<n; k++) {
      if (k != j) {
        for (i=0; i<n; i++) {
          if (i != j) {
            ret[i][k] = ret[i][k] - ret[i][j]*ret[j][k];
          }
        }
        ret[j][k] = -hr*ret[j][k];
      }
    }

  } /* END: for (j=0; j<n; j++) */

  /* Column interchange */
  for (i=0; i<n; i++) {
    for (k=0; k<n; k++) hv[p[k]] = ret[i][k];
    for (k=0; k<n; k++) ret[i][k] = hv[k];
  }

  R_Free(p);
  *error = 0;

} /* END: matrixInverse */

/* Function to compute the inverse of a matrix */
int mat_inv(double **cov, int n, double **inv)
{
  double cc, a, b, c, d;
  int error;

  switch (n) {
    case 0:
      Rprintf("\nERROR: dimension of covariance matrix is 0\n");
    case 1:
      a = cov[0][0];
      if (fabs(a) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      inv[0][0] = 1.0/a; 
      break;
    case 2:
      a  = cov[0][0];
      b  = cov[0][1];
      c  = cov[1][0];
      d  = cov[1][1];
      cc = a*d - b*c;
      if (fabs(cc) < ALMOST_ZERO) return(ERROR_SINGULAR_MATRIX);
      cc = 1.0/cc;
      inv[0][0] = d*cc;
      inv[0][1] = -b*cc;
      inv[1][0] = -c*cc;
      inv[1][1] = a*cc;
      break;
    default:
      matrixInverse(cov, n, inv, &error);
      if (error) return(error);
      break; 
  } /* END: switch */

  return(0);

} /* END: mat_inv */


